--[[

This is the constructor for dynamic 
programs that depend only on one 
argument.  It constructs a memo table, 
and then it applies the given operator 
o to cache entries in the memo table 
when they are not yet present.  Repeated 
calls for the same argument retrieve the 
cached value directly from the memo table.  



--]]

function recursively_memoize_oper(o) 
   local memo_table = {} 
   local function fn(arg) 
      local cached = memo_table[arg] 
      if cached then return cached 
      else 
	 local new_val = o(arg,fn) 
	 memo_table[arg] = new_val 
	 return new_val
      end 
   end
   return fn
end

--[[


Given a sequence s, return an 
operator that takes as its arguments 
an integer i and an "approximation" g
to the recursive calls for the optimal value.  

Look in dynamic.lua in this directory, 
at the definition of weighted_interval_sched,
for an example of how to write a function 
that returns an operator.  

--]]

--[[

*Here* fill in a blank to define the recurrence relation 
that your operator will have.  Use the instances in CLRS 
as a guide for what this should be like.  

--]]

function mnas_oper(s) 
   local function f(i,g) 

      -- Fill in the body of the operator here.
      -- This is the knowledge extension operator 
      -- that gives new answers when some old 
      -- answer's are known, and are furnished by g().

   end 
   return f 
end 

function mnas_max(s) 
   return recursively_memoize_oper(mnas_oper(s))(#s)
end 

--[[

The previous function computes the largest 
sum that a non-adjacent subsequence can achieve.
Now write a function that will print out a 
subsequence that will achieve this maximum.  
That means, you have to say which slots were 
chosen, not just what sum they lead to.

You can do this with the same operator mnas_oper, 
by using the idea on p 396 in the CLRS textbook.  
You will also find examples in the solution-printing 
routines in dynamic.lua and dynamic_lcs.lua.  

Here, you want to construct and return a Lua 
array that contains the entries you selected for the 
maximum.  

--]]

function mnas_seq(s) 
   local f = recursively_memoize_oper(mnas_oper(s))

   -- Fill in this blank.  

end 


s1 = {3,5,3,7,8,2,1,5,6}
s2 = {3,5,3,8,7,12,1,5,12}
s3 = {3,5,3,8,7,2,1,5,6,3,5,7,8,6}
s4 = {3,5,3,8,7,2,1,5,6,50,3,7,100}
s5 = {246, 216, 169, 42, 127, 49, 25, 233, 201, 25, 
      48, 95, 229, 183, 290, 196, 104, 236, 201, 162} 

-- This time, instead of using the knowledge extension operator
-- directly, program it iteratively to fill in an array from left to
-- right. 

function mnas_max_make_table_iter(s) 
   local result = {} 

   -- code to fill the entries of result left to right 
   -- goes here 

   return result 

end 

function mnas_max_seq_iter(s) 
   local memo_table = mnas_max_make_table_iter(s)

   -- write code here to recompute the *sequence* 
   -- of indices you want to use, rather than just  
   -- the value of their sum.  

   -- Work right to left in the memo_table.  
   -- Each time the value is better than the
   -- one to the left, you're using that value 
   -- after the earlier part of the past.  Otherwise, 
   -- just print the earlier part of the path.  

   -- Look at the CLRS examples in Ch 15, for instance p 395.

end 

   
      